from __future__ import with_statement
from __future__ import absolute_import
from __future__ import print_function
import sys


try:
    import numpy
    import reflex
    from pipeline_product import PipelineProduct
    import pipeline_display
    import reflex_plot_widgets

    from matplotlib import gridspec, pylab, pyplot, transforms
    import pdb  # for debugging
    from collections import defaultdict  # to make dictionary of lists
    import_success = True
    
except ImportError:
    import_success = False
    print("Error importing modules pyfits, wx, matplotlib, numpy")

# Median absolute deviation function; used to scale the images
def MAD(x):
    x=numpy.array(x)
    return numpy.median(numpy.abs(x-numpy.median(x)))


def paragraph(text, width=None):
    """ wrap text string into paragraph
       text:  text to format, removes leading space and newlines
       width: if not None, wraps text, not recommended for tooltips as
              they are wrapped by wxWidgets by default
    """
    import textwrap
    if width is None:
        return textwrap.dedent(text).replace('\n', ' ').strip()
    else:
        return textwrap.fill(textwrap.dedent(text), width=width)


class DataPlotterManager(object):
    """
    This class must be added to the PipelineInteractiveApp with setPlotManager
    It must have following member functions which will be called by the app:
     - setInteractiveParameters(self)
     - readFitsData(self, fitsFiles):
     - addSubplots(self, figure):
     - plotProductsGraphics(self, figure, canvas)
    Following members are optional:
     - setWindowHelp(self)
     - setWindowTitle(self)
    """

    # static members
    
    recipe_name = "vircam_science_process"
    stk_cat = "JITTERED_IMAGE_SCI"  #  pawprints/stacked image

    def setWindowTitle(self):
        return self.recipe_name+"_interactive"

    def setInteractiveParameters(self):
        """
        This function specifies which are the parameters that should be presented
        in the window to be edited.  Note that the parameter has to also be in the
        in_sop port (otherwise it won't appear in the window). The descriptions are
        used to show a tooltip. They should match one to one with the parameter
        list.
        """

        # Only selected Recipe parameters are shown because list is too long
        # The selection shown ones likely to be wanted to be changed by user
        return [
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="destripe",
                                   group="vircam_science_process", 
                                   description="Destripe image?. [TRUE]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="skyalgo",
                                   group="vircam_science_process", 
                                   description="Sky subtraction algorithm. <master | none | pawsky_mask | pawsky_mask_pre | tilesky | tilesky_minus | auto | pawsky_minus> [auto]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="minphotom",
                                   group="vircam_science_process", 
                                   description="Minimum number of stars for photometry solution. [1]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="cdssearch",
                                   group="vircam_science_process", 
                                   description="Use CDS for 2MASS standards? [FALSE]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="stk_cat_ipix",
                                   group="vircam_science_process", 
                                   description="Minimum pixel area for each detected object. [5]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="stk_cat_thresh",
                                   group="vircam_science_process", 
                                   description="Detection threshold in sigma above sky. [2.0]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="stk_cat_icrowd",
                                   group="vircam_science_process", 
                                   description="Use deblending?. [TRUE]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="stk_cat_rcore",
                                   group="vircam_science_process", 
                                   description="Value of Rcore in pixels. [3.0]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="stk_cat_nbsize",
                                   group="vircam_science_process", 
                                   description="Background smoothing box size. [64]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="stk_lthr",
                                   group="vircam_science_process", 
                                   description="Low rejection threshold. [5.0]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="stk_hthr",
                                   group="vircam_science_process", 
                                   description="Upper rejection threshold. [5.0]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="stk_method",
                                   group="vircam_science_process", 
                                   description="Stacking method. <nearest | linear> [linear]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="stk_seeing",
                                   group="vircam_science_process", 
                                   description="Weight by seeing?. [FALSE]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="stk_fast",
                                   group="vircam_science_process", 
                                   description="Use fast stacking?. <fast | slow | auto> [auto]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="stk_nfst",
                                   group="vircam_science_process", 
                                   description="Nframes to stack in fast mode [30]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="stk_filt",
                                   group="vircam_science_process2", 
                                   description="Filter stripe pattern? [TRUE]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="psm_ipix",
                                   group="vircam_science_process2", 
                                   description="Minimum pixel area for each detected object. [5]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="psm_niter",
                                   group="vircam_science_process2", 
                                   description="Number of iterations in pawsky mask. [5]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="psm_thresh",
                                   group="vircam_science_process2", 
                                   description="Detection threshold in sigma above sky. [2.0]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="psm_nbsize",
                                   group="vircam_science_process2", 
                                   description="Background smoothing box size. [50]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="psm_smkern",
                                   group="vircam_science_process2", 
                                   description="Smoothing kernel size (pixels). [2.0]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="cacheloc",
                                   group="vircam_science_process2", 
                                   description="Location of standard star cache [.]"),
            reflex.RecipeParameter(recipe=self.recipe_name, displayName="magerrcut",
                                   group="vircam_science_process2", 
                                   description="Magnitude error cut [100.0]")
            ]

    def readFitsData(self, fitsFiles):
        """
        This function should be used to read and organize the raw fits files
        produced by the recipes.
        It receives as input a list of reflex.FitsFiles
        """

        # frames is a dict of keyword/list pairs where elements of list are PipelineProducts
        #  it contains all FITS files in the input parameter list

#        self.stk_frames = defaultdict(list)
        self.stk_frames = []

        for f in fitsFiles:
            if (f.category == self.stk_cat):
                self.stk_frames.append(PipelineProduct(f))

        if (len(self.stk_frames)) > 0:
            
            self.stk_img_found = True
            self.n_paws = len(self.stk_frames)
            self.cur_paw = 0
            self.cur_chip_group = 0
            self.n_extn = len(self.stk_frames[0].hdulist())-1  # number of extensions, assumed to be same for all fitsFiles

            # Don't read the individual stacks/pawprints images in here 
            #   for memory/performance reasons. Read them in below as needed.
            #   There could be a lot of frames and user may not want to see all of them.

            # re-define eso-rex's pipeline_display plotting functions to enable callbacks
            self._add_subplots = self._add_subplots
            self._plot = self._data_plot

            # Define radio button options
            self.left_opts = {'Pawprint':0} 
            self.mid_opts = list(["Go to\nnext pawprint"])
            self.right_opts = list(["Go to next\nset of chips"])

            # Set the initial radio button selections (value 1 for left, 0 for mid)
            self.left_label = [key for key, value in self.left_opts.items() if value == 0][0]

        else:
            # Set the plotting functions to NODATA ones
            self._add_subplots = self._add_nodata_subplots
            self._plot = self._nodata_plot

    def addSubplots(self, figure):
        self._add_subplots(figure)

    def plotProductsGraphics(self):
        self._plot()

    def plotWidgets(self) :
        widgets = list()

        # Radio buttons 
        # Only show them if at least one stack is found
        if (self.stk_img_found == True):
            # pull out the keys from the dict() of button options sorted by value
            left_labels = [key for key,value in sorted(self.left_opts.items(),key= lambda k: k[1])]
            self.radiobutton_left = reflex_plot_widgets.InteractiveRadioButtons(self.axradiobutton_left, 
                                                                                self.setRadioCallback_left, 
                                                                                left_labels,
                                                                                self.left_opts.get(self.left_label), 
                                                                                title='Show: ')
            widgets.append(self.radiobutton_left)

            self.radiobutton_mid = reflex_plot_widgets.InteractiveRadioButtons(self.axradiobutton_mid, 
                                                                               self.setRadioCallback_mid, 
                                                                               self.mid_opts,
                                                                               0, title='')
            widgets.append(self.radiobutton_mid)

            self.radiobutton_right = reflex_plot_widgets.InteractiveRadioButtons(self.axradiobutton_right, 
                                                                                 self.setRadioCallback_right, 
                                                                                 self.right_opts,
                                                                                 0, title='')
            widgets.append(self.radiobutton_right)

            # Adjust size of button boxes and font size of labels
#            for i in range(len(widgets)):
#                pos = widgets[i].rbuttons.ax.get_position()
#                widgets[i].rbuttons.ax.set_position(transforms.Bbox([[pos.x0,pos.y0-0.01],[pos.x1, 0.97]]  ) )
#                for j in range(len(widgets[i].rbuttons.labels)):
#                    widgets[i].rbuttons.labels[j].set_fontsize(11)

        return widgets

    def setRadioCallback_left(self, label) :

        # Only do something if user changes the button
        if (label != self.left_label):
            self.left_label = label
            self._plot()

    def setRadioCallback_mid(self, label) :

        # advance (or wrap) which pawprint to look at
        self.cur_paw += 1
        if (self.cur_paw == (self.n_paws)):
            self.cur_paw = 0
        
        self._plot()

    def setRadioCallback_right(self, label) :

        # advance (or wrap) chip group
        self.cur_chip_group += 1
        if (self.cur_chip_group == (self.n_extn/4)):
            self.cur_chip_group = 0
        
        self._plot()

    def _add_subplots(self, figure):

        self.img_plot = []
        gs = gridspec.GridSpec(7, 4)
        self.img_plot.append(figure.add_subplot(gs[1:4,0:2]))
        self.img_plot.append(figure.add_subplot(gs[1:4,2:4]))
        self.img_plot.append(figure.add_subplot(gs[4:8,0:2]))
        self.img_plot.append(figure.add_subplot(gs[4:8,2:4]))
        self.axradiobutton_left = figure.add_subplot(gs[0,0])
        self.axradiobutton_mid = figure.add_subplot(gs[0,1])
        self.axradiobutton_right = figure.add_subplot(gs[0,2])
            
    def _data_plot(self):

        # Get filter name from PHU of first pawprint and assume its same for the rest
        try:
            filt_name = self.stk_frames[0].all_hdu[0].header['HIERARCH ESO INS FILT1 NAME']
        except:
            filt_name = ''
            
        for i in range(self.cur_chip_group*4,(self.cur_chip_group+1)*4,):
            chip_name = self.stk_frames[self.cur_paw].all_hdu[i+1].header['EXTNAME']

            imgdisp = pipeline_display.ImageDisplay()
            imgdisp.setAspect('equal')
            self.img_plot[i%4].cla()

            if ( ((i%4) == 0) or ((i%4)==1) ): 
                pylab.setp(self.img_plot[i%4].get_xticklabels(), visible = False)
                imgdisp.setLabels('','Y')
            else:
                pylab.setp(self.img_plot[i%4].get_xticklabels(), visible = True)
                imgdisp.setLabels('X','Y')

            if (self.left_opts[self.left_label] == 0):
                title = "Pawprint {} {}/{} Chip: {}".format(filt_name,self.cur_paw+1,self.n_paws,chip_name)
                # Try reading image
                try:
                    temp = self.stk_frames[self.cur_paw]
                    temp.readImage(i+1)

                    # Set z-limit using 1 iteration of sigma clipping using MED and MAD
                    #temp_image = temp_image[numpy.isfinite(temp_image)]
                    #med = numpy.median(temp_image)
                    #mad = MAD(temp_image)
                    #temp_image = temp_image[numpy.abs(temp_image-med) < 3*1.48*mad]
                    #new_med = numpy.median(temp_image)
                    #new_mad = MAD(temp_image)
                    # Set limits to median-1sigma, median+3sigma ?
                    #imgdisp.z_lim = new_med-1.48*new_mad, new_med+3*1.48*new_mad

                    imgdisp.display(self.img_plot[i%4], title, "Pawprint:\n " + 
                                    temp.fits_file.name, temp.image)
                except IndexError:
                    self.img_plot[i%4].set_axis_off()
                    text_nodata = "No image found for this chip/extension."
                    self.img_plot[i%4].text(0.1, 0.5, text_nodata, color='#11557c',
                                            fontsize=18, ha='left', va='center', alpha=1.0,
                                            transform = self.img_plot[i%4].transAxes)
                    self.img_plot[i%4].tooltip = 'No data found'
                    continue  # go to next extension

    def _add_nodata_subplots(self, figure):
        self.img_plot = figure.add_subplot(1,1,1)

    def _nodata_plot(self):
        # could be moved to reflex library?
        self.img_plot.set_axis_off()
        text_nodata = "Data not found. Input files should contain this" \
                       " type:\n %s" % self.stk_cat
        self.img_plot.text(0.1, 0.6, text_nodata, color='#11557c',
                      fontsize=18, ha='left', va='center', alpha=1.0)
        self.img_plot.tooltip = 'No data found'


    def setWindowHelp(self):
      help_text = """
This is an interactive window which help asses the quality of the execution of a recipe.
"""
      return help_text


#This is the 'main' function
if __name__ == '__main__':
    from reflex_interactive_app import PipelineInteractiveApp

    # Create interactive application
    interactive_app = PipelineInteractiveApp(enable_init_sop=True)

    # get inputs from the command line
    interactive_app.parse_args()

    #Check if import failed or not
    if not import_success:
        interactive_app.setEnableGUI(False)

    #Open the interactive window if enabled
    if interactive_app.isGUIEnabled():
        #Get the specific functions for this window
        dataPlotManager = DataPlotterManager()

        interactive_app.setPlotManager(dataPlotManager)
        interactive_app.showGUI()
    else:
        interactive_app.set_continue_mode()

    #Print outputs. This is parsed by the Reflex python actor to
    #get the results. Do not remove
    interactive_app.print_outputs()
    sys.exit()
